@@ -2,7 +2,7 @@
 /*
  * Ruby Cairo Binding
  *
- * Copyright 2005-2014 Kouhei Sutou <kou@cozmixng.org>
+ * Copyright 2005-2017 Kouhei Sutou <kou@cozmixng.org>
  * Copyright 2014 Patrick Hanevold <patrick.hanevold@gmail.com>
  * Copyright 2005 Øyvind Kolås <pippin@freedesktop.org>
  * Copyright 2004-2005 MenTaLguY <mental@rydia.com>
@@ -45,16 +45,7 @@
 #endif
 
 #ifdef CAIRO_HAS_QUARTZ_SURFACE
-#  ifndef HAVE_TYPE_ENUM_RUBY_VALUE_TYPE
-enum ruby_value_type {
-  RUBY_T_DATA = T_DATA
-};
-#  endif
-#  undef T_DATA
-#  include <cairo-quartz.h>
-#  define T_DATA RUBY_T_DATA
-#  define RB_CAIRO_HAS_QUARTZ_SURFACE
-#  ifdef RB_CAIRO_HAS_QUARTZ_IMAGE_SURFACE_TYPE
+#  ifdef CAIRO_HAS_QUARTZ_IMAGE_SURFACE
 #    define RB_CAIRO_HAS_QUARTZ_IMAGE_SURFACE
 #  endif
 #endif
@@ -129,8 +120,8 @@ cr_paper_to_size_in_points (VALUE paper_description, VALUE *width, VALUE *height
   *height = RARRAY_PTR (size)[1];
 }
 
-static inline void
-cr_surface_check_status (cairo_surface_t *surface)
+void
+rb_cairo_surface_check_status (cairo_surface_t *surface)
 {
   rb_cairo_check_status (cairo_surface_status (surface));
 }
@@ -358,6 +349,24 @@ cr_surface_xml_supported_p (VALUE klass)
 #endif
 }
 
+static void
+rb_cairo_surface_adjust_memory_usage (cairo_surface_t *surface,
+                                      cairo_bool_t new)
+{
+#ifdef HAVE_RB_GC_ADJUST_MEMORY_USAGE
+  if (cairo_surface_get_type (surface) == CAIRO_SURFACE_TYPE_IMAGE)
+    {
+      ssize_t memory_usage_diff;
+      memory_usage_diff =
+        cairo_image_surface_get_stride (surface) *
+        cairo_image_surface_get_height (surface);
+      if (!new)
+        memory_usage_diff = -memory_usage_diff;
+      rb_gc_adjust_memory_usage (memory_usage_diff);
+    }
+#endif
+}
+
 /* constructor/de-constructor */
 cairo_surface_t *
 rb_cairo_surface_from_ruby_object (VALUE obj)
@@ -390,8 +399,12 @@ cr_surface_free (void *ptr)
 {
   cairo_surface_t *surface = ptr;
 
-  if (surface)
-    cairo_surface_destroy (surface);
+  if (!surface)
+    return;
+
+  rb_cairo_surface_adjust_memory_usage (surface, CR_FALSE);
+
+  cairo_surface_destroy (surface);
 }
 
 VALUE
@@ -402,6 +415,7 @@ rb_cairo_surface_to_ruby_object (cairo_surface_t *surface)
       VALUE klass;
       klass = cr_surface_get_klass (surface);
       cairo_surface_reference (surface);
+      rb_cairo_surface_adjust_memory_usage (surface, CR_TRUE);
       return Data_Wrap_Struct (klass, NULL, cr_surface_free, surface);
     }
   else
@@ -451,6 +465,18 @@ cr_surface_destroy (VALUE self)
   return self;
 }
 
+static VALUE
+cr_surface_get_reference_count (VALUE self)
+{
+  cairo_surface_t *surface;
+  unsigned int reference_count;
+
+  surface = _SELF;
+  reference_count = cairo_surface_get_reference_count (surface);
+
+  return UINT2NUM (reference_count);
+}
+
 static VALUE
 cr_surface_finish (VALUE self)
 {
@@ -466,13 +492,13 @@ cr_surface_finish (VALUE self)
 
   if (closure && !NIL_P (closure->error))
     rb_exc_raise (closure->error);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
 
   return self;
 }
 
-static void
-yield_and_finish (VALUE self)
+void
+rb_cairo__surface_yield_and_finish (VALUE self)
 {
   cairo_surface_t *surface;
 
@@ -513,7 +539,7 @@ cr_surface_create_similar (int argc, VALUE *argv, VALUE self)
 
   similar_surface = cairo_surface_create_similar (surface, content,
                                                   width, height);
-  cr_surface_check_status (similar_surface);
+  rb_cairo_surface_check_status (similar_surface);
   return CRSURFACE2RVAL_WITH_DESTROY (similar_surface);
 }
 
@@ -551,7 +577,7 @@ cr_surface_create_similar_image (int argc, VALUE *argv, VALUE self)
 
   similar_image = cairo_surface_create_similar_image (surface, format,
                                                       width, height);
-  cr_surface_check_status (similar_image);
+  rb_cairo_surface_check_status (similar_image);
   return CRSURFACE2RVAL_WITH_DESTROY (similar_image);
 }
 
@@ -589,7 +615,7 @@ cr_surface_map_to_image (int argc, VALUE *argv, VALUE self)
     }
 
   mapped_image = cairo_surface_map_to_image (surface, extents);
-  cr_surface_check_status (mapped_image);
+  rb_cairo_surface_check_status (mapped_image);
   return CRSURFACE2RVAL_WITH_DESTROY (mapped_image);
 }
 
@@ -626,7 +652,7 @@ cr_surface_create_sub_rectangle_surface (VALUE self, VALUE x, VALUE y,
                                                 NUM2DBL (y),
                                                 NUM2DBL (width),
                                                 NUM2INT (height));
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   rb_surface = CRSURFACE2RVAL_WITH_DESTROY (surface);
   if (rb_block_given_p ())
     return rb_ensure (rb_yield, rb_surface,
@@ -758,7 +784,7 @@ cr_surface_get_font_options (VALUE self)
 {
   cairo_font_options_t *options = cairo_font_options_create();
   cairo_surface_get_font_options (_SELF, options);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   rb_cairo_check_status (cairo_font_options_status (options));
   return CRFONTOPTIONS2RVAL (options);
 }
@@ -767,7 +793,7 @@ static VALUE
 cr_surface_flush (VALUE self)
 {
   cairo_surface_flush (_SELF);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return self;
 }
 
@@ -805,7 +831,7 @@ cr_surface_mark_dirty (int argc, VALUE *argv, VALUE self)
                 rb_cairo__inspect (args));
     }
 
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return self;
 }
 
@@ -815,7 +841,7 @@ cr_surface_set_device_offset (VALUE self, VALUE x_offset, VALUE y_offset)
   cairo_surface_set_device_offset (_SELF,
                                    NUM2DBL (x_offset),
                                    NUM2DBL (y_offset));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return self;
 }
 
@@ -825,10 +851,31 @@ cr_surface_get_device_offset (VALUE self)
   double x_offset, y_offset;
 
   cairo_surface_get_device_offset (_SELF, &x_offset, &y_offset);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return rb_ary_new3 (2, rb_float_new (x_offset), rb_float_new (y_offset));
 }
 
+#if CAIRO_CHECK_VERSION(1, 14, 0)
+static VALUE
+cr_surface_set_device_scale (VALUE self, VALUE x_scale, VALUE y_scale)
+{
+  cairo_surface_set_device_scale (_SELF,
+                                  NUM2DBL (x_scale),
+                                  NUM2DBL (y_scale));
+  rb_cairo_surface_check_status (_SELF);
+  return self;
+}
+
+static VALUE
+cr_surface_get_device_scale (VALUE self)
+{
+  double x_scale, y_scale;
+
+  cairo_surface_get_device_scale (_SELF, &x_scale, &y_scale);
+  return rb_ary_new3 (2, rb_float_new (x_scale), rb_float_new (y_scale));
+}
+#endif
+
 static VALUE
 cr_surface_set_fallback_resolution (VALUE self,
                                     VALUE x_pixels_per_inch,
@@ -837,7 +884,7 @@ cr_surface_set_fallback_resolution (VALUE self,
   cairo_surface_set_fallback_resolution (_SELF,
                                          NUM2DBL (x_pixels_per_inch),
                                          NUM2DBL (y_pixels_per_inch));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return self;
 }
 
@@ -850,7 +897,7 @@ cr_surface_get_fallback_resolution (VALUE self)
   cairo_surface_get_fallback_resolution (_SELF,
                                          &x_pixels_per_inch,
                                          &y_pixels_per_inch);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return rb_ary_new3 (2,
                       rb_float_new (x_pixels_per_inch),
                       rb_float_new (y_pixels_per_inch));
@@ -863,7 +910,7 @@ static VALUE
 cr_surface_copy_page (VALUE self)
 {
   cairo_surface_copy_page (_SELF);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return self;
 }
 
@@ -871,7 +918,7 @@ static VALUE
 cr_surface_show_page (VALUE self)
 {
   cairo_surface_show_page (_SELF);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return self;
 }
 #endif
@@ -912,7 +959,7 @@ cr_image_surface_create_from_png_generic (VALUE klass, VALUE target)
   else
     surface = cr_image_surface_create_from_png (target);
 
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   rb_surface = cr_surface_allocate (klass);
   DATA_PTR (rb_surface) = surface;
   return rb_surface;
@@ -970,10 +1017,11 @@ cr_image_surface_initialize (int argc, VALUE *argv, VALUE self)
               "(data, format, width, height, stride)): %s",
               rb_cairo__inspect (rb_ary_new3 (4, arg1, arg2, arg3, arg4)));
 
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
+  rb_cairo_surface_adjust_memory_usage (surface, CR_TRUE);
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 
@@ -1047,6 +1095,12 @@ cr_ ## type ## _surface_initialize (int argc, VALUE *argv, VALUE self)  \
   width_in_points = NUM2DBL (rb_width_in_points);                       \
   height_in_points = NUM2DBL (rb_height_in_points);                     \
                                                                         \
+  if (rb_respond_to (target, rb_cairo__io_id_to_path) &&                \
+      !rb_respond_to (target, rb_cairo__io_id_to_io))                   \
+    {                                                                   \
+      target = rb_funcall (target, rb_cairo__io_id_to_path, 0);         \
+    }                                                                   \
+                                                                        \
   if (rb_respond_to (target, rb_cairo__io_id_write))                    \
     {                                                                   \
       rb_cairo__io_callback_closure_t *closure;                         \
@@ -1076,16 +1130,18 @@ cr_ ## type ## _surface_initialize (int argc, VALUE *argv, VALUE self)  \
     }                                                                   \
   else                                                                  \
     {                                                                   \
-      surface =                                                         \
-        cairo_ ## type ## _surface_create (StringValueCStr (target),    \
-                                           width_in_points,             \
-                                           height_in_points);           \
+      const char *filename = NULL;                                      \
+      if (!NIL_P (target))                                              \
+        filename = StringValueCStr (target);                            \
+      surface = cairo_ ## type ## _surface_create (filename,            \
+                                                   width_in_points,     \
+                                                   height_in_points);   \
     }                                                                   \
                                                                         \
-  cr_surface_check_status (surface);                                    \
+  rb_cairo_surface_check_status (surface);                              \
   DATA_PTR (self) = surface;                                            \
   if (rb_block_given_p ())                                              \
-    yield_and_finish (self);                                            \
+    rb_cairo__surface_yield_and_finish (self);                          \
   return Qnil;                                                          \
 }
 
@@ -1115,7 +1171,7 @@ cr_ ## type ## _surface_set_size (int argc, VALUE *argv, VALUE self)    \
   cairo_ ## type ## _surface_set_size (_SELF,                           \
                                        NUM2DBL (width_in_points),       \
                                        NUM2DBL (height_in_points));     \
-  cr_surface_check_status (_SELF);                                      \
+  rb_cairo_surface_check_status (_SELF);                                      \
   return Qnil;                                                          \
 }
 
@@ -1129,7 +1185,112 @@ static VALUE
 cr_pdf_surface_restrict_to_version (VALUE self, VALUE version)
 {
   cairo_pdf_surface_restrict_to_version (_SELF, RVAL2CRPDFVERSION (version));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
+  return Qnil;
+}
+#  endif
+
+#  if CAIRO_CHECK_VERSION(1, 15, 4)
+static VALUE
+cr_pdf_surface_add_outline (VALUE self,
+                            VALUE rb_parent_id,
+                            VALUE rb_name,
+                            VALUE rb_destination,
+                            VALUE rb_flags)
+{
+  cairo_surface_t *surface;
+  int parent_id;
+  const char *name;
+  const char *destination;
+  cairo_pdf_outline_flags_t flags;
+  int id;
+
+  surface = _SELF;
+  if (NIL_P (rb_parent_id))
+    parent_id = 0;
+  else
+    parent_id = NUM2INT (rb_parent_id);
+  name = RVAL2CSTR (rb_name);
+  destination = RVAL2CSTR (rb_destination);
+  if (NIL_P (rb_flags))
+    flags = 0;
+  else
+    flags = RVAL2CRPDFOUTLINEFLAGS (rb_flags);
+  id = cairo_pdf_surface_add_outline (surface,
+                                      parent_id,
+                                      name,
+                                      destination,
+                                      flags);
+  rb_cairo_surface_check_status (surface);
+
+  if (id == 0)
+    return Qnil;
+  else
+    return INT2NUM (id);
+}
+
+static VALUE
+cr_pdf_surface_set_metadata (VALUE self,
+                             VALUE rb_metadata,
+                             VALUE rb_value)
+{
+  cairo_surface_t *surface;
+  cairo_pdf_metadata_t metadata;
+  const char *value;
+
+  surface = _SELF;
+  metadata = RVAL2CRPDFMETADATA (rb_metadata);
+  switch (metadata)
+    {
+    case CAIRO_PDF_METADATA_CREATE_DATE:
+    case CAIRO_PDF_METADATA_MOD_DATE:
+      if (rb_cairo__is_kind_of (rb_value, rb_cTime))
+        {
+          ID id_iso8601;
+          CONST_ID(id_iso8601, "iso8601");
+          rb_value = rb_funcall (rb_value, id_iso8601, 0);
+        }
+      break;
+    default:
+      break;
+    }
+  value = RVAL2CSTR (rb_value);
+  cairo_pdf_surface_set_metadata (surface, metadata, value);
+  rb_cairo_surface_check_status (surface);
+
+  return Qnil;
+}
+
+static VALUE
+cr_pdf_surface_set_page_label (VALUE self,
+                               VALUE rb_label)
+{
+  cairo_surface_t *surface;
+  const char *label;
+
+  surface = _SELF;
+  label = RVAL2CSTR (rb_label);
+  cairo_pdf_surface_set_page_label (surface, label);
+  rb_cairo_surface_check_status (surface);
+
+  return Qnil;
+}
+
+static VALUE
+cr_pdf_surface_set_thumbnail_size (VALUE self,
+                                   VALUE rb_width,
+                                   VALUE rb_height)
+{
+  cairo_surface_t *surface;
+  int width;
+  int height;
+
+  surface = _SELF;
+  width = NUM2INT (rb_width);
+  height = NUM2INT (rb_height);
+  cairo_pdf_surface_set_thumbnail_size (surface, width, height);
+  rb_cairo_surface_check_status (surface);
+
   return Qnil;
 }
 #  endif
@@ -1144,7 +1305,7 @@ static VALUE
 cr_ps_surface_dsc_comment (VALUE self, VALUE comment)
 {
   cairo_ps_surface_dsc_comment (_SELF, StringValueCStr (comment));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return Qnil;
 }
 
@@ -1152,7 +1313,7 @@ static VALUE
 cr_ps_surface_dsc_begin_setup (VALUE self)
 {
   cairo_ps_surface_dsc_begin_setup (_SELF);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   if (rb_block_given_p ())
     return rb_yield (self);
   else
@@ -1163,7 +1324,7 @@ static VALUE
 cr_ps_surface_dsc_begin_page_setup (VALUE self)
 {
   cairo_ps_surface_dsc_begin_page_setup (_SELF);
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   if (rb_block_given_p ())
     return rb_yield (self);
   else
@@ -1175,7 +1336,7 @@ static VALUE
 cr_ps_surface_restrict_to_level (VALUE self, VALUE level)
 {
   cairo_ps_surface_restrict_to_level (_SELF, RVAL2CRPSLEVEL (level));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return Qnil;
 }
 
@@ -1189,143 +1350,15 @@ static VALUE
 cr_ps_surface_set_eps (VALUE self, VALUE eps)
 {
   cairo_ps_surface_set_eps (_SELF, RTEST (eps));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return Qnil;
 }
 #  endif
 #endif
 
-#ifdef RB_CAIRO_HAS_QUARTZ_SURFACE
-/* Quartz-surface functions */
-#include <objc/objc-runtime.h>
-
-BOOL rbobj_to_nsobj (VALUE obj, id* nsobj);
-VALUE ocid_to_rbobj (VALUE context_obj, id ocid);
-
-static VALUE
-cr_quartz_surface_initialize (int argc, VALUE *argv, VALUE self)
-{
-  id objc_object = nil;
-  CGContextRef context;
-  unsigned int width, height;
-  cairo_surface_t *surface = NULL;
-  cairo_format_t format = CAIRO_FORMAT_ARGB32;
-  VALUE arg1, arg2, arg3, rb_width, rb_height;
-#ifdef HAVE_RUBY_COCOA
-  static VALUE rb_cOSXCGContextRef = Qnil;
-#endif
-  static VALUE rb_cFFIPointer = Qnil;
-
-  rb_scan_args (argc, argv, "21", &arg1, &arg2, &arg3);
-
-  if (argc == 2)
-    {
-      rb_width = arg1;
-      rb_height = arg2;
-    }
-  else
-    {
-      switch (TYPE (arg1))
-        {
-        case T_NIL:
-          break;
-        case T_STRING:
-        case T_SYMBOL:
-        case T_FIXNUM:
-          format = RVAL2CRFORMAT (arg1);
-          break;
-        default:
-#ifdef HAVE_RUBY_COCOA
-          if (NIL_P (rb_cOSXCGContextRef))
-            rb_cOSXCGContextRef =
-              rb_const_get (rb_const_get (rb_cObject, rb_intern ("OSX")),
-                            rb_intern ("CGContextRef"));
-#endif
-
-          if (NIL_P (rb_cFFIPointer))
-            rb_cFFIPointer =
-              rb_const_get (rb_const_get (rb_cObject, rb_intern ("FFI")),
-                            rb_intern ("Pointer"));
-
-#ifdef HAVE_RUBY_COCOA
-          if (RTEST (rb_obj_is_kind_of (arg1, rb_cOSXCGContextRef)))
-            {
-              rbobj_to_nsobj (arg1, &objc_object);
-            }
-          else
-#endif
-            {
-              if (RTEST (rb_obj_is_kind_of (arg1, rb_cFFIPointer)))
-                {
-                  VALUE rb_objc_pointer;
-                  rb_objc_pointer = rb_funcall (arg1,
-                                                rb_intern ("address"),
-                                                0);
-                  objc_object = NUM2ULONG (rb_objc_pointer);
-                }
-              else
-                {
-                  rb_raise (rb_eArgError,
-                            "invalid argument (expect "
-                            "(width, height), "
-                            "(format, width, height), "
-                            "(cg_context, width, height) or "
-                            "(ffi_pointer, width, height)): %s",
-                            rb_cairo__inspect (rb_ary_new3 (3, arg1, arg2, arg3)));
-                }
-            }
-          break;
-        }
-
-      rb_width = arg2;
-      rb_height = arg3;
-    }
-
-  width = NUM2UINT (rb_width);
-  height = NUM2UINT (rb_height);
-
-  if (objc_object == nil)
-    {
-      surface = cairo_quartz_surface_create (format, width, height);
-    }
-  else
-    {
-      context = (CGContextRef)objc_object;
-      surface =
-        cairo_quartz_surface_create_for_cg_context (context, width, height);
-    }
-
-  cr_surface_check_status (surface);
-  DATA_PTR (self) = surface;
-  if (rb_block_given_p ())
-    yield_and_finish (self);
-  return Qnil;
-}
-
-static VALUE
-cr_quartz_surface_get_cg_context (VALUE self)
-{
-  CGContextRef context;
-  id objc_object;
-
-  context = cairo_quartz_surface_get_cg_context (_SELF);
-  objc_object = (id)context;
-  return ocid_to_rbobj (Qnil, objc_object);
-}
-#endif
-
 #ifdef CAIRO_HAS_WIN32_SURFACE
 /* Win32 surface functions */
 
-/* from dl/dl.h (ruby 1.9) */
-#  if SIZEOF_LONG == SIZEOF_VOIDP
-#    define PTR2NUM(x)   (ULONG2NUM((unsigned long)(x)))
-#    define NUM2PTR(x)   ((void *)(NUM2ULONG(x)))
-#  else
-#    define PTR2NUM(x)   (ULL2NUM((unsigned long long)(x)))
-#    define NUM2PTR(x)   ((void *)(NUM2ULL(x)))
-#  endif
-
 static VALUE
 cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
 {
@@ -1342,11 +1375,34 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
       surface = cairo_win32_surface_create (NUM2PTR (hdc));
       break;
     case 2:
-      width = arg1;
-      height = arg2;
-      surface = cairo_win32_surface_create_with_dib (CAIRO_FORMAT_ARGB32,
-                                                     NUM2INT (width),
-                                                     NUM2INT (height));
+#  if CAIRO_CHECK_VERSION(1, 11, 4)
+#    define CAIRO_FORMAT_MAX CAIRO_FORMAT_RGB30
+#  else
+#    define CAIRO_FORMAT_MAX CAIRO_FORMAT_RGB16_565
+#  endif
+      if (CAIRO_FORMAT_ARGB32 <= NUM2INT (arg2) &&
+          NUM2INT (arg2) <= CAIRO_FORMAT_MAX)
+        {
+#  if CAIRO_CHECK_VERSION(1, 15, 2)
+          hdc = arg1;
+          format = arg2;
+          surface =
+            cairo_win32_surface_create_with_format (NUM2PTR (hdc),
+                                                    RVAL2CRFORMAT (format));
+#  else
+          rb_raise (rb_eArgError,
+                    "Cairo::Win32Surface.new(hdc, format) "
+                    "is available since cairo >= 1.15.2");
+#  endif
+        }
+      else
+        {
+          width = arg1;
+          height = arg2;
+          surface = cairo_win32_surface_create_with_dib (CAIRO_FORMAT_ARGB32,
+                                                         NUM2INT (width),
+                                                         NUM2INT (height));
+        }
       break;
     case 3:
       if (NIL_P (arg1) ||
@@ -1387,7 +1443,7 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
         format = arg2;
         width = arg3;
         height = arg4;
-        win32_hdc = NIL_P (hdc) ? NULL : (HDC) NUM2UINT (hdc);
+        win32_hdc = NIL_P (hdc) ? NULL : (HDC) (UINT_PTR) NUM2UINT (hdc);
         surface = cairo_win32_surface_create_with_ddb (win32_hdc,
                                                        RVAL2CRFORMAT (format),
                                                        NUM2INT (width),
@@ -1403,10 +1459,10 @@ cr_win32_surface_initialize (int argc, VALUE *argv, VALUE self)
 
   if (!surface)
     rb_cairo_check_status (CAIRO_STATUS_INVALID_FORMAT);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 
@@ -1431,7 +1487,7 @@ cr_win32_surface_get_image (VALUE self)
   surface = cairo_win32_surface_get_image (_SELF);
   if (!surface)
     return Qnil;
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   return CRSURFACE2RVAL (surface);
 }
 #  endif
@@ -1445,9 +1501,30 @@ static VALUE
 cr_svg_surface_restrict_to_version (VALUE self, VALUE version)
 {
   cairo_svg_surface_restrict_to_version (_SELF, RVAL2CRSVGVERSION (version));
-  cr_surface_check_status (_SELF);
+  rb_cairo_surface_check_status (_SELF);
   return Qnil;
 }
+
+#  if CAIRO_CHECK_VERSION(1, 15, 10)
+static VALUE
+cr_svg_surface_get_document_unit (VALUE self)
+{
+  cairo_svg_unit_t unit;
+  unit = cairo_svg_surface_get_document_unit (_SELF);
+  rb_cairo_surface_check_status (_SELF);
+  return INT2NUM (unit);
+}
+
+static VALUE
+cr_svg_surface_set_document_unit (VALUE self, VALUE rb_unit)
+{
+  cairo_svg_unit_t unit;
+  unit = RVAL2CRSVGUNIT (rb_unit);
+  cairo_svg_surface_set_document_unit (_SELF, unit);
+  rb_cairo_surface_check_status (_SELF);
+  return Qnil;
+}
+#  endif
 #endif
 
 #ifdef RB_CAIRO_HAS_WIN32_PRINTING_SURFACE
@@ -1458,42 +1535,14 @@ cr_win32_printing_surface_initialize (VALUE self, VALUE hdc)
   cairo_surface_t *surface = NULL;
 
   surface = cairo_win32_printing_surface_create (NUM2PTR (hdc));
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 #endif
 
-#ifdef RB_CAIRO_HAS_QUARTZ_IMAGE_SURFACE
-/* Quartz image surface functions */
-static VALUE
-cr_quartz_image_surface_initialize (VALUE self, VALUE image_surface)
-{
-  cairo_surface_t *surface;
-
-  surface = cairo_quartz_image_surface_create (RVAL2CRSURFACE (image_surface));
-  cr_surface_check_status (surface);
-  DATA_PTR (self) = surface;
-  if (rb_block_given_p ())
-    yield_and_finish (self);
-  return Qnil;
-}
-
-static VALUE
-cr_quartz_image_surface_get_image (VALUE self)
-{
-  cairo_surface_t *surface;
-
-  surface = cairo_quartz_image_surface_get_image (_SELF);
-  if (!surface)
-    return Qnil;
-  cr_surface_check_status (surface);
-  return CRSURFACE2RVAL (surface);
-}
-#endif
-
 #ifdef CAIRO_HAS_SCRIPT_SURFACE
 /* script surface functions */
 static VALUE
@@ -1541,10 +1590,10 @@ cr_script_surface_initialize (int argc, VALUE *argv, VALUE self)
   else
     surface = cairo_script_surface_create (device, content, width, height);
 
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 #endif
@@ -1596,10 +1645,10 @@ cr_recording_surface_initialize (int argc, VALUE *argv, VALUE self)
     }
 
   surface = cairo_recording_surface_create (content, &extents);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 
@@ -1611,7 +1660,7 @@ cr_recording_surface_get_ink_extents (VALUE self)
 
   surface = _SELF;
   cairo_recording_surface_ink_extents (surface, &x, &y, &width, &height);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   return rb_ary_new3 (4,
                       rb_float_new (x), rb_float_new (y),
                       rb_float_new (width), rb_float_new (height));
@@ -1626,7 +1675,7 @@ cr_recording_surface_get_extents (VALUE self)
 
   surface = _SELF;
   cairo_recording_surface_get_extents (surface, &extents);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   return rb_ary_new3 (4,
                       rb_float_new (extents.x),
                       rb_float_new (extents.y),
@@ -1673,10 +1722,10 @@ cr_gl_surface_initialize (int argc, VALUE *argv, VALUE self)
 
   surface = cairo_gl_surface_create (device, content, width, height);
 
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 
@@ -1720,10 +1769,10 @@ cr_gl_texture_surface_initialize (int argc, VALUE *argv, VALUE self)
                                                  width,
                                                  height);
 
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 
@@ -1734,7 +1783,7 @@ cr_gl_surface_set_size (VALUE self, VALUE width, VALUE height)
 
   surface = _SELF;
   cairo_gl_surface_set_size (surface, NUM2INT (width), NUM2INT (height));
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   return Qnil;
 }
 
@@ -1757,7 +1806,7 @@ cr_gl_surface_swap_buffers (VALUE self)
 
   surface = _SELF;
   cairo_gl_surface_swapbuffers (surface);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   return Qnil;
 }
 #endif
@@ -1770,11 +1819,11 @@ cr_tee_surface_initialize (VALUE self, VALUE master)
   cairo_surface_t *surface = NULL;
 
   surface = cairo_tee_surface_create (RVAL2CRSURFACE (master));
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   rb_iv_set (self, "surfaces", rb_ary_new3 (1, master));
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 
@@ -1785,7 +1834,7 @@ cr_tee_surface_add (VALUE self, VALUE target)
 
   surface = _SELF;
   cairo_tee_surface_add (surface, RVAL2CRSURFACE (target));
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   rb_ary_push (rb_iv_get (self, "surfaces"), target);
   return Qnil;
 }
@@ -1821,7 +1870,7 @@ cr_tee_surface_remove (VALUE self, VALUE target_or_index)
       target = cairo_tee_surface_index (surface, NUM2INT (index));
     }
   cairo_tee_surface_remove (surface, target);
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
 
   rb_surfaces = rb_iv_get (self, "surfaces");
   for (i = 0; i < RARRAY_LEN (rb_surfaces); i++)
@@ -1849,8 +1898,8 @@ cr_tee_surface_array_reference (VALUE self, VALUE index)
   surface = _SELF;
   index = rb_Integer (index);
   target = cairo_tee_surface_index (surface, NUM2UINT (index));
-  cr_surface_check_status (surface);
-  cr_surface_check_status (target);
+  rb_cairo_surface_check_status (surface);
+  rb_cairo_surface_check_status (target);
   return CRSURFACE2RVAL (target);
 }
 #endif
@@ -1892,10 +1941,10 @@ cr_xml_surface_initialize (int argc, VALUE *argv, VALUE self)
 
   surface = cairo_xml_surface_create (device, content, width, height);
 
-  cr_surface_check_status (surface);
+  rb_cairo_surface_check_status (surface);
   DATA_PTR (self) = surface;
   if (rb_block_given_p ())
-    yield_and_finish (self);
+    rb_cairo__surface_yield_and_finish (self);
   return Qnil;
 }
 #endif
@@ -1982,6 +2031,8 @@ Init_cairo_surface (void)
                     cr_surface_get_device, 0);
 #endif
   rb_define_method (rb_cCairo_Surface, "destroy", cr_surface_destroy, 0);
+  rb_define_method (rb_cCairo_Surface, "reference_count",
+                    cr_surface_get_reference_count, 0);
   rb_define_method (rb_cCairo_Surface, "finish", cr_surface_finish, 0);
   rb_define_method (rb_cCairo_Surface, "content", cr_surface_get_content, 0);
 
@@ -2005,6 +2056,12 @@ Init_cairo_surface (void)
                     cr_surface_set_device_offset, 2);
   rb_define_method (rb_cCairo_Surface, "device_offset",
                     cr_surface_get_device_offset, 0);
+#if CAIRO_CHECK_VERSION(1, 14, 0)
+  rb_define_method (rb_cCairo_Surface, "set_device_scale",
+                    cr_surface_set_device_scale, 2);
+  rb_define_method (rb_cCairo_Surface, "device_scale",
+                    cr_surface_get_device_scale, 0);
+#endif
   rb_define_method (rb_cCairo_Surface, "set_fallback_resolution",
                     cr_surface_set_fallback_resolution, 2);
 #if CAIRO_CHECK_VERSION(1, 7, 2)
@@ -2063,6 +2120,26 @@ Init_cairo_surface (void)
                     cr_pdf_surface_restrict_to_version, 1);
 #  endif
 
+#  if CAIRO_CHECK_VERSION(1, 15, 4)
+  {
+    VALUE rb_mCairo_PDFOutline;
+
+    rb_mCairo_PDFOutline =
+      rb_define_module_under (rb_mCairo, "PDFOutline");
+    rb_define_const (rb_mCairo_PDFOutline, "ROOT",
+                     INT2NUM (CAIRO_PDF_OUTLINE_ROOT));
+
+    rb_define_method (rb_cCairo_PDFSurface, "add_outline",
+                      cr_pdf_surface_add_outline, 4);
+    rb_define_method (rb_cCairo_PDFSurface, "set_metadata",
+                      cr_pdf_surface_set_metadata, 2);
+    rb_define_method (rb_cCairo_PDFSurface, "set_page_label",
+                      cr_pdf_surface_set_page_label, 1);
+    rb_define_method (rb_cCairo_PDFSurface, "set_thumbnail_size",
+                      cr_pdf_surface_set_thumbnail_size, 2);
+  }
+#  endif
+
   RB_CAIRO_DEF_SETTERS (rb_cCairo_PDFSurface);
 #endif
 
@@ -2101,12 +2178,6 @@ Init_cairo_surface (void)
   /* Quartz surface */
   rb_cCairo_QuartzSurface =
     rb_define_class_under (rb_mCairo, "QuartzSurface", rb_cCairo_Surface);
-#ifdef RB_CAIRO_HAS_QUARTZ_SURFACE
-  rb_define_method (rb_cCairo_QuartzSurface, "initialize",
-                    cr_quartz_surface_initialize, -1);
-  rb_define_method (rb_cCairo_QuartzSurface, "cg_context",
-                    cr_quartz_surface_get_cg_context, 0);
-#endif
 
   /* Win32 surface */
   rb_cCairo_Win32Surface =
@@ -2130,6 +2201,12 @@ Init_cairo_surface (void)
                     cr_svg_surface_initialize, -1);
   rb_define_method (rb_cCairo_SVGSurface, "restrict_to_version",
                     cr_svg_surface_restrict_to_version, 1);
+#  if CAIRO_CHECK_VERSION(1, 15, 10)
+  rb_define_method (rb_cCairo_SVGSurface, "document_unit",
+                    cr_svg_surface_get_document_unit, 0);
+  rb_define_method (rb_cCairo_SVGSurface, "set_document_unit",
+                    cr_svg_surface_set_document_unit, 1);
+#  endif
 
   RB_CAIRO_DEF_SETTERS (rb_cCairo_SVGSurface);
 #endif
@@ -2147,12 +2224,6 @@ Init_cairo_surface (void)
   /* Quartz image surface */
   rb_cCairo_QuartzImageSurface =
     rb_define_class_under (rb_mCairo, "QuartzImageSurface", rb_cCairo_Surface);
-#ifdef RB_CAIRO_HAS_QUARTZ_IMAGE_SURFACE
-  rb_define_method (rb_cCairo_QuartzImageSurface, "initialize",
-                    cr_quartz_image_surface_initialize, 1);
-  rb_define_method (rb_cCairo_QuartzImageSurface, "image",
-                    cr_quartz_image_surface_get_image, 0);
-#endif
 
   /* script surface */
   rb_cCairo_ScriptSurface =
