@@ -1,5 +1,7 @@
 require 'cairo'
+require 'poppler'
 require 'stringio'
+require 'test/unit/notify'
 
 module CairoTestUtils
   private
@@ -9,12 +11,20 @@ module CairoTestUtils
     end
   end
 
-  def only_win32
-    omit("Only for Win32 platform") unless win32
+  def only_windows
+    omit("Only for Windows platform") unless windows?
   end
 
-  def win32?
-    /cygwin|mingw|mswin32|bccwin32/.match(RUBY_PLATFORM) ? true : false
+  def only_not_windows
+    omit("Only for not Windows platform") if windows?
+  end
+
+  def windows?
+    /cygwin|mingw|mswin|bccwin/ === RUBY_PLATFORM
+  end
+
+  def quartz?
+    Cairo::FontFace.quartz_supported?
   end
 
   def only_device(name)
@@ -36,4 +46,8 @@ module CairoTestUtils
       omit("Only for #{name} device available")
     end
   end
+
+  def fixture_path(*components)
+    File.join(__dir__, "fixture", *components)
+  end
 end
